@@ -10,3 +10,5 @@ from .registers import RegisterModelAdmin
 from .roles import PortfolioRoleAdmin
 from .transactions import DividendAdmin, FeesAdmin, TradeAdmin
 from .reconciliations import AccountReconciliationAdmin
+from .orders import OrderProposalAdmin
+from .rebalancing import RebalancingModelAdmin, RebalancerAdmin
