@@ -400,8 +400,8 @@ class DistributionFilter(wb_filters.FilterSet):
 
     group_by = wb_filters.ChoiceFilter(
         label="Group By",
-        choices=AssetPositionGroupBy.choices(),
-        initial=AssetPositionGroupBy.INDUSTRY.name,
+        choices=AssetPositionGroupBy.choices,
+        initial=AssetPositionGroupBy.INDUSTRY.value,
         method="fake_filter",
         clearable=False,
         required=True,
@@ -427,6 +427,14 @@ class DistributionFilter(wb_filters.FilterSet):
         endpoint=ClassificationGroup.get_representation_endpoint(),
         value_key=ClassificationGroup.get_representation_value_key(),
         label_key=ClassificationGroup.get_representation_label_key(),
+        depends_on=[{"field": "group_by", "options": {"activates_on": [AssetPositionGroupBy.INDUSTRY.value]}}],
+    )
+
+    group_by_classification_height = wb_filters.NumberFilter(
+        method="fake_filter",
+        label="Classification Height",
+        initial=0,
+        depends_on=[{"field": "group_by", "options": {"activates_on": [AssetPositionGroupBy.INDUSTRY.value]}}],
     )
 
     class Meta:
