@@ -4,6 +4,7 @@ from decimal import Decimal
 
 from celery import shared_task
 from django.contrib import admin
+from django.contrib.contenttypes.fields import GenericRelation
 from django.contrib.postgres.constraints import ExclusionConstraint
 from django.contrib.postgres.fields import DateRangeField, RangeOperators
 from django.db import models
@@ -33,6 +34,7 @@ from wbcore.utils.enum import ChoiceEnum
 from wbcrm.models.accounts import Account
 from wbfdm.models.instruments.instrument_prices import InstrumentPrice
 from wbfdm.models.instruments.instruments import InstrumentManager, InstrumentType
+from wbreport.models import Report
 
 from wbportfolio.models.portfolio_relationship import InstrumentPortfolioThroughModel
 
@@ -198,6 +200,8 @@ class FeeProductPercentage(models.Model):
 
 
 class Product(PMSInstrumentAbstractModel):
+    reports = GenericRelation(Report)
+
     share_price = models.PositiveIntegerField(
         default=100,
         verbose_name="Share Price",
@@ -344,6 +348,11 @@ class Product(PMSInstrumentAbstractModel):
 
         self.is_managed = True
 
+    def save(self, *args, **kwargs):
+        super().save(*args, **kwargs)
+        if self.delisted_date and self.delisted_date <= date.today():
+            self.reports.update(is_active=False)
+
     def get_title(self):
         if self.parent:
             return f"{self.parent.name} ({self.name})"
