@@ -5,12 +5,12 @@ from unittest.mock import patch
 
 import pandas as pd
 import pytest
-from django.contrib.contenttypes.models import ContentType
 from django.db.models import F, Sum
 from django.forms.models import model_to_dict
 from faker import Faker
 from pandas.tseries.offsets import BDay
 from psycopg.types.range import DateRange
+from wbcore.contrib.currency.factories import CurrencyFactory
 from wbcore.contrib.geography.factories import CountryFactory
 
 from wbportfolio.models import (
@@ -20,9 +20,8 @@ from wbportfolio.models import (
     PortfolioPortfolioThroughModel,
     Trade,
 )
-from wbportfolio.models.asset import AssetPositionIterator
 
-from ...models.portfolio import get_returns, update_portfolio_after_investable_universe
+from ...models.portfolio import update_portfolio_after_investable_universe
 from .utils import PortfolioTestMixin
 
 fake = Faker()
@@ -108,11 +107,10 @@ class TestPortfolioModel(PortfolioTestMixin):
         )
         assert portfolio.get_geographical_breakdown(weekday).shape[0] == 2
 
-    def test_get_currency_exposure(self, portfolio, asset_position_factory, currency_factory, equity_factory, weekday):
+    def test_get_currency_exposure(self, portfolio, asset_position_factory, equity_factory, weekday):
         a1 = asset_position_factory.create(
             portfolio=portfolio,
-            underlying_instrument=equity_factory.create(),
-            currency=currency_factory.create(),
+            underlying_instrument=equity_factory.create(currency=CurrencyFactory.create()),
             date=weekday,
         )
         asset_position_factory.create(
@@ -120,8 +118,7 @@ class TestPortfolioModel(PortfolioTestMixin):
         )
         asset_position_factory.create(
             portfolio=portfolio,
-            underlying_instrument=equity_factory.create(),
-            currency=currency_factory.create(),
+            underlying_instrument=equity_factory.create(currency=CurrencyFactory.create()),
             date=weekday,
         )
         assert portfolio.get_currency_exposure(weekday).shape[0] == 2
@@ -255,16 +252,18 @@ class TestPortfolioModel(PortfolioTestMixin):
 
     @patch.object(Portfolio, "estimate_net_asset_values", autospec=True)
     def test_change_at_date(self, mock_estimate_net_asset_values, asset_position_factory, portfolio, weekday):
-        asset_position_factory.create_batch(10, portfolio=portfolio, date=weekday)
+        a1 = asset_position_factory.create(portfolio=portfolio, date=weekday, weighting=Decimal("0.1"))
+        a2 = asset_position_factory.create(portfolio=portfolio, date=weekday, weighting=Decimal("0.3"))
+        a3 = asset_position_factory.create(portfolio=portfolio, date=weekday, weighting=Decimal("0.5"))
 
-        portfolio.change_at_date(weekday, recompute_weighting=True)
+        portfolio.change_at_date(weekday, fix_quantization=True)
 
-        # test that change at date normalize the weighting
-        total_value = AssetPosition.objects.aggregate(s=Sum("total_value_fx_portfolio"))["s"]
-        for pos in AssetPosition.objects.all():
-            assert float(pos.weighting) == pytest.approx(float(pos.total_value_fx_portfolio / total_value), rel=1e-2)
+        cash_pos = portfolio.assets.get(date=weekday, portfolio=portfolio, underlying_quote=portfolio.cash_component)
+        assert cash_pos.weighting == Decimal("1.0") - (a1.weighting + a2.weighting + a3.weighting)
 
-        mock_estimate_net_asset_values.assert_called_once_with(portfolio, (weekday + BDay(1)).date(), weights=None)
+        mock_estimate_net_asset_values.assert_called_once_with(
+            portfolio, (weekday + BDay(1)).date(), analytic_portfolio=None
+        )
 
     @patch.object(Portfolio, "compute_lookthrough", autospec=True)
     def test_change_at_date_with_dependent_portfolio(
@@ -282,7 +281,10 @@ class TestPortfolioModel(PortfolioTestMixin):
         dependent_portfolio.depends_on.add(base_portfolio)
         base_portfolio.change_at_date(weekday)
 
-        mock_compute_lookthrough.assert_called_once_with(dependent_portfolio, weekday)
+        mock_compute_lookthrough.assert_called_once_with(
+            dependent_portfolio,
+            weekday,
+        )
 
     def test_is_active_at_date(
         self,
@@ -527,8 +529,9 @@ class TestPortfolioModel(PortfolioTestMixin):
         assert res1
 
     def test_get_total_asset_under_management(
-        self, portfolio, customer_trade_factory, instrument_factory, instrument_price_factory, weekday
+        self, portfolio_factory, customer_trade_factory, instrument_factory, instrument_price_factory, weekday
     ):
+        portfolio = portfolio_factory.create()
         i1 = instrument_factory.create()
         i2 = instrument_factory.create()
         previous_day = (weekday - BDay(5)).date()
@@ -998,20 +1001,18 @@ class TestPortfolioModel(PortfolioTestMixin):
         fx_portfolio = currency_fx_rates_factory.create(currency=portfolio.currency, date=weekday)
         fx_instrument = currency_fx_rates_factory.create(currency=instrument.currency, date=weekday)
         instrument_id: int = instrument.id
-        weights = {instrument_id: random.random()}
-        positions = AssetPositionIterator(
-            portfolio, prices={weekday: {instrument_id: p.net_value}}, infer_underlying_quote_price=False
-        )
-        positions.add((weekday, weights))
+        weights = {instrument_id: Decimal(random.random())}
+        portfolio.builder.prices = {weekday: {instrument_id: p.net_value}}
+        portfolio.builder.add((weekday, weights), infer_underlying_quote_price=False)
 
-        res = list(positions)
+        res = list(portfolio.builder.get_positions())
         a = res[0]
         assert len(res) == 1
         assert a.date == weekday
         assert a.underlying_quote == instrument
         assert a.underlying_quote_price == None
         assert a.initial_price == p.net_value
-        assert a.weighting == weights[instrument.id]
+        assert a.weighting == pytest.approx(weights[instrument.id], abs=Decimal(10e-6))
         assert a.currency_fx_rate_portfolio_to_usd == fx_portfolio
         assert a.currency_fx_rate_instrument_to_usd == fx_instrument
 
@@ -1047,32 +1048,34 @@ class TestPortfolioModel(PortfolioTestMixin):
         asset_position_factory.create(date=weekday, portfolio=portfolio, underlying_instrument=i2, weighting=0.3)
 
         rebalancer_factory.create(portfolio=portfolio, frequency="RRULE:FREQ=DAILY;", activation_date=rebalancing_date)
-        positions, rebalancing_trade_proposal = portfolio.drift_weights(weekday, (rebalancing_date + BDay(1)).date())
-        assert rebalancing_trade_proposal.trade_date == rebalancing_date
-        assert rebalancing_trade_proposal.status == "SUBMIT"
-        assert set(positions.get_weights().keys()) == {
-            middle_date,
-        }, "Drifting weight with a non automatic rebalancer stops the iteration"
+        portfolio.load_builder_returns(weekday, rebalancing_date)
+        gen = portfolio.drift_weights(weekday, (rebalancing_date + BDay(1)).date(), stop_at_rebalancing=True)
+        assert next(gen)[0] == middle_date, "Drifting weight with a non automatic rebalancer stops the iteration"
+        try:
+            next(gen)
+            assert False, "the next iteration should stop and return the rebalancing"
+        except StopIteration as e:
+            rebalancing_order_proposal = e.value
+            assert rebalancing_order_proposal.trade_date == rebalancing_date
+            assert rebalancing_order_proposal.status == "APPROVED"
 
-        # we expect a equally rebalancing (default) so both trades needs to be created
-        t1 = rebalancing_trade_proposal.trades.get(transaction_date=rebalancing_date, underlying_instrument=i1)
-        t2 = rebalancing_trade_proposal.trades.get(transaction_date=rebalancing_date, underlying_instrument=i2)
-        assert t1._target_weight == Decimal("0.5")
-        assert t2._target_weight == Decimal("0.5")
+            # we expect a equally rebalancing (default) so both orders needs to be created
+            orders = rebalancing_order_proposal.get_orders()
+            t1 = orders.get(value_date=rebalancing_date, underlying_instrument=i1)
+            t2 = orders.get(value_date=rebalancing_date, underlying_instrument=i2)
+            assert t1._target_weight == Decimal("0.5")
+            assert t2._target_weight == Decimal("0.5")
 
-        # we approve the rebalancing trade proposal
-        assert rebalancing_trade_proposal.status == "SUBMIT"
-        rebalancing_trade_proposal.approve(no_replay=True)
-        rebalancing_trade_proposal.save()
+            # we approve the rebalancing order proposal
+            assert rebalancing_order_proposal.status == "APPROVED"
+            rebalancing_order_proposal.apply(replay=False)
+            rebalancing_order_proposal.save()
 
         # check that the rebalancing was applied and position reflect that
         assert portfolio.assets.get(date=rebalancing_date, underlying_instrument=i1).weighting == Decimal("0.5")
         assert portfolio.assets.get(date=rebalancing_date, underlying_instrument=i2).weighting == Decimal("0.5")
 
-    @patch("wbportfolio.models.portfolio.compute_metrics_as_task.delay")
-    def test_bulk_create_positions(
-        self, mock_compute_metrics, portfolio, weekday, asset_position_factory, instrument_factory
-    ):
+    def test_bulk_create_positions(self, portfolio, weekday, asset_position_factory, instrument_factory):
         portfolio.is_manageable = False
         portfolio.save()
         i1 = instrument_factory.create()
@@ -1081,21 +1084,17 @@ class TestPortfolioModel(PortfolioTestMixin):
         a1 = asset_position_factory.build(date=weekday, portfolio=portfolio, underlying_instrument=i1)
 
         # check initial creation
-        portfolio.bulk_create_positions(AssetPositionIterator(portfolio).add([a1]))
+        portfolio.builder.add([a1]).bulk_create_positions(fix_quantization=False, compute_metrics=True)
         assert AssetPosition.objects.get(portfolio=portfolio, date=weekday).weighting == a1.weighting
         assert AssetPosition.objects.get(portfolio=portfolio, date=weekday).underlying_instrument == i1
 
-        mock_compute_metrics.assert_called_once_with(
-            weekday, basket_id=portfolio.id, basket_content_type_id=ContentType.objects.get_for_model(Portfolio).id
-        )
-
-        # check that if we change key value, an already exising position will be updated accordingly
+        # check that if we change key value, an already existing position will be updated accordingly
         a1.weighting = Decimal(0.5)
-        portfolio.bulk_create_positions(AssetPositionIterator(portfolio).add([a1]))
+        portfolio.builder.add([a1]).bulk_create_positions(fix_quantization=False)
         assert AssetPosition.objects.get(portfolio=portfolio, date=weekday).weighting == Decimal(0.5)
 
         a2 = asset_position_factory.build(date=weekday, portfolio=portfolio, underlying_instrument=i2)
-        portfolio.bulk_create_positions(AssetPositionIterator(portfolio).add([a2]))
+        portfolio.builder.add([a2]).bulk_create_positions(fix_quantization=False)
         assert (
             AssetPosition.objects.get(portfolio=portfolio, date=weekday, underlying_instrument=i1).weighting
             == a1.weighting
@@ -1106,7 +1105,7 @@ class TestPortfolioModel(PortfolioTestMixin):
         )
 
         a3 = asset_position_factory.build(date=weekday, portfolio=portfolio, underlying_instrument=i3)
-        portfolio.bulk_create_positions(AssetPositionIterator(portfolio).add([a3]), delete_leftovers=True)
+        portfolio.builder.add([a3]).bulk_create_positions(delete_leftovers=True, fix_quantization=False)
         assert AssetPosition.objects.get(portfolio=portfolio, date=weekday).weighting == a3.weighting
         assert AssetPosition.objects.get(portfolio=portfolio, date=weekday).underlying_instrument == i3
 
@@ -1125,33 +1124,6 @@ class TestPortfolioModel(PortfolioTestMixin):
         res = list(Portfolio.objects.all().to_dependency_iterator(weekday))
         assert res == [index_portfolio, dependency_portfolio, dependant_portfolio, undependant_portfolio]
 
-    def test_get_returns(self, instrument_factory, instrument_price_factory, asset_position_factory, portfolio):
-        v1 = date(2025, 1, 1)
-        v2 = date(2025, 1, 2)
-        v3 = date(2025, 1, 3)
-
-        i1 = instrument_factory.create()
-        i2 = instrument_factory.create()
-
-        i11 = instrument_price_factory.create(date=v1, instrument=i1)
-        i12 = instrument_price_factory.create(date=v2, instrument=i1)
-        i13 = instrument_price_factory.create(date=v3, instrument=i1)
-        asset_position_factory.create(date=v1, portfolio=portfolio, underlying_instrument=i1)
-        asset_position_factory.create(date=v3, portfolio=portfolio, underlying_instrument=i2)
-        i11.refresh_from_db()
-        i12.refresh_from_db()
-        i13.refresh_from_db()
-        returns = get_returns([i1.id, i2.id], from_date=v1, to_date=v3)
-
-        expected_returns = pd.DataFrame(
-            [[i12.net_value / i11.net_value - 1, 0.0], [i13.net_value / i12.net_value - 1, 0.0]],
-            index=[v2, v3],
-            columns=[i1.id, i2.id],
-            dtype="float64",
-        )
-        expected_returns.index = pd.to_datetime(expected_returns.index)
-        pd.testing.assert_frame_equal(returns, expected_returns, check_names=False, check_freq=False, atol=1e-6)
-
     @patch.object(Portfolio, "compute_lookthrough", autospec=True)
     def test_handle_controlling_portfolio_change_at_date(self, mock_compute_lookthrough, weekday, portfolio_factory):
         primary_portfolio = portfolio_factory.create(only_weighting=True)
@@ -1164,3 +1136,37 @@ class TestPortfolioModel(PortfolioTestMixin):
 
         primary_portfolio.handle_controlling_portfolio_change_at_date(weekday)
         mock_compute_lookthrough.assert_called_once_with(lookthrough_portfolio, weekday)
+
+    def test_get_model_portfolio_relationships(self, portfolio_factory, asset_position_factory, weekday):
+        model_portfolio = portfolio_factory.create()
+        model_index = model_portfolio.get_or_create_index()
+        dependent_portfolio = portfolio_factory.create()
+        re1 = PortfolioPortfolioThroughModel.objects.create(
+            portfolio=dependent_portfolio,
+            dependency_portfolio=model_portfolio,
+            type=PortfolioPortfolioThroughModel.Type.MODEL,
+        )
+
+        assert set(model_portfolio.get_model_portfolio_relationships(weekday)) == {
+            re1,
+        }
+        parent_portfolio = portfolio_factory.create()
+        child_portfolio = portfolio_factory.create()
+        re2 = PortfolioPortfolioThroughModel.objects.create(
+            portfolio=child_portfolio,
+            dependency_portfolio=parent_portfolio,
+            type=PortfolioPortfolioThroughModel.Type.MODEL,
+        )
+        assert set(model_portfolio.get_model_portfolio_relationships(weekday)) == {
+            re1,
+        }  # child portfolio is not considered in the tree because there is no position yet
+
+        asset_position_factory.create(portfolio=parent_portfolio, underlying_instrument=model_index, date=weekday)
+        assert set(model_portfolio.get_model_portfolio_relationships(weekday)) == {re1, re2}
+
+        dependent_portfolio.is_active = False  # disable this portfolio
+        dependent_portfolio.deletion_datetime = weekday - timedelta(days=1)
+        dependent_portfolio.save()
+        assert set(model_portfolio.get_model_portfolio_relationships(weekday)) == {
+            re2,
+        }
