@@ -2,23 +2,15 @@ from contextlib import suppress
 from datetime import date, datetime
 
 import pandas as pd
-import plotly.graph_objects as go
 from django.db.models import Exists, F, OuterRef, Q, Sum
 from django.utils.dateparse import parse_date
 from django.utils.functional import cached_property
-from plotly.subplots import make_subplots
 from wbcore import viewsets
-from wbcore.contrib.currency.models import Currency, CurrencyFXRates
+from wbcore.contrib.currency.models import CurrencyFXRates
 from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
-from wbcore.filters import DjangoFilterBackend
 from wbcore.pandas import fields as pf
 from wbcore.permissions.permissions import InternalUserPermissionMixin
 from wbcore.serializers import decorator
-from wbcore.utils.date import get_date_interval_from_request
-from wbcore.utils.figures import (
-    get_default_timeserie_figure,
-    get_hovertemplate_timeserie,
-)
 from wbcore.utils.strings import format_number
 from wbfdm.contrib.metric.viewsets.mixins import InstrumentMetricMixin
 from wbfdm.models import Instrument
@@ -27,11 +19,8 @@ from wbportfolio.filters import (
     AssetPositionFilter,
     AssetPositionInstrumentFilter,
     AssetPositionPortfolioFilter,
-    AssetPositionUnderlyingInstrumentChartFilter,
     CashPositionPortfolioFilterSet,
-    CompositionContributionChartFilter,
     CompositionModelPortfolioPandasFilter,
-    ContributionChartFilter,
 )
 from wbportfolio.import_export.resources.assets import AssetPositionResource
 from wbportfolio.metric.backends.portfolio_base import (
@@ -69,16 +58,12 @@ from .configs import (
     AssetPositionPortfolioEndpointConfig,
     AssetPositionPortfolioTitleConfig,
     AssetPositionTitleConfig,
-    AssetPositionUnderlyingInstrumentChartEndpointConfig,
-    AssetPositionUnderlyingInstrumentChartTitleConfig,
     CashPositionPortfolioDisplayConfig,
     CashPositionPortfolioEndpointConfig,
     CashPositionPortfolioTitleConfig,
     CompositionModelPortfolioPandasDisplayConfig,
     CompositionModelPortfolioPandasEndpointConfig,
     CompositionModelPortfolioPandasTitleConfig,
-    ContributorPortfolioChartEndpointConfig,
-    ContributorPortfolioChartTitleConfig,
 )
 from .mixins import UserPortfolioRequestPermissionMixin
 
@@ -133,7 +118,7 @@ class AssetPositionModelViewSet(
             weighting = queryset.aggregate(s=Sum(F("weighting")))["s"]
             aggregates.update(
                 {
-                    "weighting": {"Σ": format_number(weighting, decimal=4)},
+                    "weighting": {"Σ": format_number(weighting, decimal=8)},
                     "total_value_fx_usd": {"Σ": format_number(total_value_fx_usd)},
                 }
             )
@@ -204,7 +189,7 @@ class AssetPositionPortfolioModelViewSet(InstrumentMetricMixin, AssetPositionMod
         total_value_fx_portfolio = queryset.aggregate(s=Sum(F("total_value_fx_portfolio")))["s"]
         aggregates = super().get_aggregates(queryset, paginated_queryset)
         aggregates["total_value_fx_portfolio"] = {"Σ": format_number(total_value_fx_portfolio)}
-        aggregates["weighting"] = {"Σ": format_number(weighting, decimal=4)}
+        aggregates["weighting"] = {"Σ": format_number(weighting, decimal=8)}
         return aggregates
 
     def get_queryset(self):
@@ -344,190 +329,6 @@ class CashPositionPortfolioPandasAPIView(
         return AssetPosition.objects.none()
 
 
-# ##### CHART VIEWS #####
-
-
-class ContributorPortfolioChartView(UserPortfolioRequestPermissionMixin, viewsets.ChartViewSet):
-    filterset_class = ContributionChartFilter
-    filter_backends = (DjangoFilterBackend,)
-    IDENTIFIER = "wbportfolio:portfolio-contributor"
-    queryset = AssetPosition.objects.all()
-
-    title_config_class = ContributorPortfolioChartTitleConfig
-    endpoint_config_class = ContributorPortfolioChartEndpointConfig
-
-    ROW_HEIGHT: int = 20
-
-    @property
-    def min_height(self):
-        if hasattr(self, "nb_rows"):
-            return self.nb_rows * self.ROW_HEIGHT
-        return "300px"
-
-    @cached_property
-    def hedged_currency(self) -> Currency | None:
-        if "hedged_currency" in self.request.GET:
-            with suppress(Currency.DoesNotExist):
-                return Currency.objects.get(pk=self.request.GET["hedged_currency"])
-
-    @cached_property
-    def show_lookthrough(self) -> bool:
-        return self.portfolio.is_composition and self.request.GET.get("show_lookthrough", "false").lower() == "true"
-
-    def get_filterset_class(self, request):
-        if self.portfolio.is_composition:
-            return CompositionContributionChartFilter
-        return ContributionChartFilter
-
-    def get_plotly(self, queryset):
-        fig = go.Figure()
-        data = []
-        if self.show_lookthrough:
-            d1, d2 = get_date_interval_from_request(self.request)
-            for _d in pd.date_range(d1, d2):
-                for pos in self.portfolio.get_lookthrough_positions(_d.date()):
-                    data.append(
-                        [
-                            pos.date,
-                            pos.initial_price,
-                            pos.initial_currency_fx_rate,
-                            pos.underlying_instrument_id,
-                            pos.weighting,
-                        ]
-                    )
-        else:
-            data = queryset.annotate_hedged_currency_fx_rate(self.hedged_currency).values_list(
-                "date", "price", "hedged_currency_fx_rate", "underlying_instrument", "weighting"
-            )
-        df = Portfolio.get_contribution_df(data).rename(columns={"group_key": "underlying_instrument"})
-        if not df.empty:
-            df = df[["contribution_total", "contribution_forex", "underlying_instrument"]].sort_values(
-                by="contribution_total", ascending=True
-            )
-
-            df["instrument_id"] = df.underlying_instrument.map(
-                dict(Instrument.objects.filter(id__in=df["underlying_instrument"]).values_list("id", "name_repr"))
-            )
-            df_forex = df[["instrument_id", "contribution_forex"]]
-            df_forex = df_forex[df_forex.contribution_forex != 0]
-
-            contribution_equity = df.contribution_total - df.contribution_forex
-
-            text_forex = df_forex.contribution_forex.apply(lambda x: f"{x:,.2%}")
-            text_equity = contribution_equity.apply(lambda x: f"{x:,.2%}")
-            setattr(self, "nb_rows", df.shape[0])
-            fig.add_trace(
-                go.Bar(
-                    y=df.instrument_id,
-                    x=contribution_equity,
-                    name="Contribution Equity",
-                    orientation="h",
-                    marker=dict(
-                        color="rgba(247,110,91,0.6)",
-                        line=dict(color="rgb(247,110,91,1.0)", width=2),
-                    ),
-                    text=text_equity.values,
-                    textposition="auto",
-                )
-            )
-            fig.add_trace(
-                go.Bar(
-                    y=df_forex.instrument_id,
-                    x=df_forex.contribution_forex,
-                    name="Contribution Forex",
-                    orientation="h",
-                    marker=dict(
-                        color="rgba(58, 71, 80, 0.6)",
-                        line=dict(color="rgba(58, 71, 80, 1.0)", width=2),
-                    ),
-                    text=text_forex.values,
-                    textposition="outside",
-                )
-            )
-            fig.update_layout(
-                barmode="relative",
-                xaxis=dict(showgrid=False, showline=False, zeroline=False, tickformat=".2%"),
-                yaxis=dict(showgrid=False, showline=False, zeroline=False, tickmode="linear"),
-                margin=dict(b=0, r=20, l=20, t=0, pad=20),
-                paper_bgcolor="rgba(0,0,0,0)",
-                plot_bgcolor="rgba(0,0,0,0)",
-                font=dict(family="roboto", size=12, color="black"),
-                bargap=0.3,
-            )
-            # fig = get_horizontal_barplot(df, x_label="contribution_total", y_label="name")
-        return fig
-
-    def parse_figure_dict(self, figure_dict: dict[str, any]) -> dict[str, any]:
-        figure_dict = super().parse_figure_dict(figure_dict)
-        figure_dict["style"]["minHeight"] = self.min_height
-        return figure_dict
-
-    def get_queryset(self):
-        if self.has_portfolio_access:
-            return super().get_queryset().filter(portfolio=self.portfolio)
-        return AssetPosition.objects.none()
-
-
-class AssetPositionUnderlyingInstrumentChartViewSet(UserPortfolioRequestPermissionMixin, viewsets.ChartViewSet):
-    IDENTIFIER = "wbportfolio:assetpositionchart"
-
-    queryset = AssetPosition.objects.all()
-
-    title_config_class = AssetPositionUnderlyingInstrumentChartTitleConfig
-    endpoint_config_class = AssetPositionUnderlyingInstrumentChartEndpointConfig
-    filterset_class = AssetPositionUnderlyingInstrumentChartFilter
-
-    def get_queryset(self):
-        return AssetPosition.objects.filter(underlying_quote__in=self.instrument.get_descendants(include_self=True))
-
-    def get_plotly(self, queryset):
-        fig = make_subplots(specs=[[{"secondary_y": True}]])
-        fig = get_default_timeserie_figure(fig)
-        if queryset.exists():
-            df_price = (pd.DataFrame(queryset.values("date", "price_fx_usd"))).groupby("date").first()
-
-            df_price = df_price.where(pd.notnull(df_price), 0)
-            fig.add_trace(
-                go.Scatter(
-                    x=df_price.index, y=df_price.price_fx_usd, mode="lines", marker_color="green", name="Price"
-                ),
-                secondary_y=False,
-            )
-
-            df_weight = pd.DataFrame(queryset.values("date", "weighting", "portfolio__name"))
-            df_weight = df_weight.where(pd.notnull(df_weight), 0)
-            df_weight = df_weight.groupby(["date", "portfolio__name"]).sum().reset_index()
-            for portfolio_name, df_tmp in df_weight.groupby("portfolio__name"):
-                fig.add_trace(
-                    go.Scatter(
-                        x=df_tmp.date,
-                        y=df_tmp.weighting,
-                        hovertemplate=get_hovertemplate_timeserie(is_percent=True),
-                        mode="lines",
-                        name=f"Allocation: {portfolio_name}",
-                    ),
-                    secondary_y=True,
-                )
-
-            # Set x-axis title
-            fig.update_xaxes(title_text="Date")
-            # Set y-axes titles
-            fig.update_yaxes(
-                title_text="<b>Price</b>",
-                secondary_y=False,
-                titlefont=dict(color="green"),
-                tickfont=dict(color="green"),
-            )
-            fig.update_yaxes(
-                title_text="<b>Portfolio Allocation (%)</b>",
-                secondary_y=True,
-                titlefont=dict(color="blue"),
-                tickfont=dict(color="blue"),
-            )
-
-        return fig
-
-
 class CompositionModelPortfolioPandasView(
     UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, ExportPandasAPIViewSet
 ):
